/*
 * ALSA Patch Bay
 *
 * Copyright (C) 2002 Robert Ham (node@users.sourceforge.net)
 *
 * You have permission to use this file under the GNU General
 * Public License, version 2 or later.  See the file COPYING
 * for the full text.
 *
 */

#include "apb.h"

static APB::DebugLevel debugLevel = APB::DEBUG5;

static void * xmalloc (size_t size) {
  void * ptr;
  ptr = malloc (size);
  if (!ptr) {
    APB::AlsaPatchBay::report (APB::ERROR, string("out of memory"));
    exit (1);
  }
  return ptr;
}

static void * xrealloc (void * mem, size_t size) {
  void * ptr;
  ptr = realloc (mem, size);
  if (!ptr) {
    APB::AlsaPatchBay::report (APB::ERROR, string("out of memory"));
    exit (1);
  }
  return ptr;
}

namespace APB {

string Misc::int2string (int i) {
  static uint size = 16;
  static char * s = NULL;
  int err = -1;

  if (!s) s = (char *) xmalloc (size * sizeof(char));

  while (err == -1) {
    err = snprintf (s, size, "%d", i);
    if (err == -1) {
      size += 16;
      s = (char *) xrealloc (s, size);
    }
  }

  return string (s);
}

bool Misc::addrCmp (const Addr& a1, const Addr& a2) {
  return a1.client == a2.client && a1.port == a2.port;
}

void AlsaPatchBay::quitButtonPushed (Fl_Widget * widget, void * data) {
  // should clean up here
  exit (0);
}

void AlsaPatchBay::refreshButtonPushed (Fl_Widget * widget, void * data) {
 _apb->refreshButtonPushed ();
}

void AlsaPatchBay::refreshButtonPushed () {
  refreshPortButtons ();
  _selectedButton = 0;
  refreshSubscriptions ();
}

void AlsaPatchBay::readButtonPushed (Fl_Widget * widget, void * data) {
  _apb->readButtonPushed ((Fl_Button *) widget, (const Addr * const) data);
}

void AlsaPatchBay::writeButtonPushed (Fl_Widget * widget, void * data) {
  _apb->writeButtonPushed ((PortButton *) widget, *((const Addr *) data));
}

void AlsaPatchBay::readButtonPushed (Fl_Button * button, const Addr * const addr) {
  if (_selectedButton) {
    if (_selectedButton == button) {
      button->labelcolor (FL_BLACK);
      button->redraw ();
      _selectedButton = 0;
      return;
    } else {
      _selectedButton->labelcolor (FL_BLACK);
      _selectedButton->redraw ();
    }
  }
  button->labelcolor (FL_RED);
  button->redraw ();
  _selectedButton = button;
  _selectedPort = *addr;
}

void AlsaPatchBay::unsubscribe (const Addr& from, const Addr& to) {
  snd_seq_port_subscribe_t *subs;
  snd_seq_port_subscribe_alloca(&subs);
  snd_seq_port_subscribe_set_sender(subs, &from);
  snd_seq_port_subscribe_set_dest(subs, &to);
  snd_seq_port_subscribe_set_queue(subs, 0);
  snd_seq_port_subscribe_set_exclusive(subs, 0);
  snd_seq_port_subscribe_set_time_update(subs, 0);
  snd_seq_port_subscribe_set_time_real(subs, 0);
  
  if (snd_seq_get_port_subscription(seq, subs) != 0) {
    reportLog (ERROR, string("Subscription ") + getAddrName(from) + " -> "
                        + getAddrName(to) + " doesn't exist :-/");
  }

  int err = snd_seq_unsubscribe_port(seq, subs);
  if (err < 0) {
    reportLog (ERROR, string ("Could not remove subscription ") +
                      getAddrName(from) + " -> "  + getAddrName(to));
    return;
  }
  reportLog (NOTICE, string ("Remove subscription ")  +
                     getAddrName(from) + " -> "  + getAddrName(to));
}

void AlsaPatchBay::refreshSubscriptions () {
  getSubscriptions ();
  _scrollPane->redraw ();
  redraw();
}

void AlsaPatchBay::writeButtonPushed (PortButton * button, const Addr& addr) {
  if (!_selectedButton) {
    bool changed = false;
    for (list<Subscription>::iterator i = _subscriptions.begin ();
         i != _subscriptions.end ();
         ++i) {
      if (Misc::addrCmp (addr, i->second)) {
        unsubscribe (i->first, addr);
        changed = true;
      }
    }
    if (changed) refreshSubscriptions ();
    return;
  }

  snd_seq_port_subscribe_t *subs;
  snd_seq_port_subscribe_alloca(&subs);
  snd_seq_port_subscribe_set_sender(subs, &_selectedPort);
  snd_seq_port_subscribe_set_dest(subs, &addr);
  snd_seq_port_subscribe_set_queue(subs, 0);
  snd_seq_port_subscribe_set_exclusive(subs, 0);
  snd_seq_port_subscribe_set_time_update(subs, 0);
  snd_seq_port_subscribe_set_time_real(subs, 0);
  if (snd_seq_get_port_subscription(seq, subs) == 0) {
    // already subscribed
    reportLog (WARNING, string("Subscription ") + getAddrName(_selectedPort) + " -> "
                        + getAddrName(addr) + " already exists");
    _selectedButton->labelcolor (FL_BLACK);
    _selectedButton->redraw ();
    _selectedButton = 0;
    return;
  }
  int err = snd_seq_subscribe_port(seq, subs);
  if (err < 0) {
    reportLog (ERROR, string("Subscription ") + getAddrName(_selectedPort) + " -> "
                      + getAddrName(addr) + " failed");
    _selectedButton->labelcolor (FL_BLACK);
    _selectedButton->redraw ();
    _selectedButton = 0;
    return;
  }

  _subscriptions.push_back (Subscription (_selectedPort, addr));
  _selectedButton->labelcolor (FL_BLACK);
  _selectedButton->redraw ();
  _selectedButton = 0;
  reportLog (NOTICE, "Created subscription " +
                      getAddrName(_selectedPort) + " -> "  + getAddrName(addr));
  redraw ();
}

PortButton * AlsaPatchBay::getReadButton (const Addr& addr) {
  return getButton (addr, _readButtons);
}

PortButton * AlsaPatchBay::getWriteButton (const Addr& addr) {
  return getButton (addr, _writeButtons);
}

PortButton * AlsaPatchBay::getButton (const Addr& addr, const list<PortButton *>& portButtons) {
  for (list<PortButton *>::const_iterator i = portButtons.begin();
       i != portButtons.end();
       ++i) {
    Addr buttonAddr = (*i)->addr();
    if (buttonAddr.client == addr.client &&
        buttonAddr.port == addr.port) return *i;
  }
  return 0;
}

void AlsaPatchBay::refreshPortButtons () {
  getPorts ();

  for (list<PortButton *>::iterator i = _readButtons.begin();
       i != _readButtons.end();
       ++i) {
    _readPack->remove (*(*i));
    delete *i;
  }
  _readButtons.clear ();

  for (list<PortButton *>::iterator i = _writeButtons.begin();
       i != _writeButtons.end();
       ++i) {
    _writePack->remove (*(*i));
    delete *i;
  }
  _writeButtons.clear ();
  
  PortButton * but;
  _readPack->begin();
  for (list<Addr>::iterator i = _readPorts.begin();
       i != _readPorts.end();
       ++i) {
    but = new PortButton (0, 0, 0, buttonHeight, *i,
      getClientName (&*i) + "::" + getPortName(&*i));
    but->callback (&readButtonPushed, &*i);
    but->align (FL_ALIGN_WRAP|FL_ALIGN_CENTER);
    _readButtons.push_back (but);
  }
  _readPack->end();

  _writePack->begin();
  for (list<Addr>::iterator i = _writePorts.begin();
       i != _writePorts.end();
       ++i) {
    but = new PortButton (0, 0, 0, buttonHeight, *i,
      getClientName (&*i) + "::" + getPortName(&*i));
    but->callback (&writeButtonPushed, &*i);
    but->align (FL_ALIGN_WRAP|FL_ALIGN_CENTER);
    _writeButtons.push_back (but);
  }
  _writePack->end();
}

void AlsaPatchBay::draw () {
  Fl_Window::draw ();
  PortButton * readButton, * writeButton;
  int bX, bY, eX, eY;

  fl_push_clip(_scrollPane->x(), _scrollPane->y(),
               _scrollPane->w(), _scrollPane->h());

  for (list<Subscription>::iterator i = _subscriptions.begin();
       i != _subscriptions.end();
       ++i) {
    readButton = getReadButton (i->first);
    writeButton = getWriteButton (i->second);
    bX = readButton->w();
    bY = readButton->y() + (readButton->h() / 2);
    eX = writeButton->x();
    eY = writeButton->y() + (writeButton->h() / 2);

    Fl_Color oldColor = fl_color ();
    fl_color (readButton->color ());
    fl_line(bX, bY, eX, eY);
    fl_color (oldColor);
  }

  fl_pop_clip ();
}

void AlsaPatchBay::fltkInit (int argc, char ** argv) {
//  size_range(300, 300);
  int vunit = (h()-buttonHeight) / 4;

  begin();
  _log = new Fl_Multiline_Output(0, (vunit * 3) + buttonHeight, w(), vunit);

  _quitButton = new Fl_Button((w())-buttonWidth,
                           0, buttonWidth, buttonHeight, "Quit");
  _quitButton->callback(&quitButtonPushed);


  _refreshButton = new Fl_Button((w())-(buttonWidth*2),
                              0, buttonWidth, buttonHeight, "Refresh");
  _refreshButton->callback(&refreshButtonPushed);

  _scrollPane = new Fl_Scroll (0, buttonHeight, w(),
                               vunit * 3);
  int packWidth = ((w()-16) / 7) * 3;
  _readPack = new Fl_Pack (0, buttonHeight,
                           packWidth, vunit * 3);
  _readPack->end();
  _writePack = new Fl_Pack (_scrollPane->w() - packWidth - 16, buttonHeight,
                            packWidth, vunit * 3);
  _writePack->end();
  _scrollPane->end();

  end();
  show(argc, argv);

}

string AlsaPatchBay::getPortName (const Addr * const addr) const {
  snd_seq_port_info_t * pinfo;
  snd_seq_port_info_alloca (&pinfo);
  int err = snd_seq_get_any_port_info (seq, addr->client, addr->port, pinfo);
  if (err) {
    string s ("queried non-existant port");
    APB_REPORT (ERROR, s);
    throw Exception (s);
  }
  return string (snd_seq_port_info_get_name (pinfo));
}

string AlsaPatchBay::getClientName (const Addr * const addr) const {
  snd_seq_client_info_t * cinfo;
  snd_seq_client_info_alloca (&cinfo);
  int err = snd_seq_get_any_client_info (seq, addr->client, cinfo);
  if (err) {
    string s ("queried non-existant client");
    APB_REPORT (ERROR, s);
    throw Exception (s);
  }
  return string (snd_seq_client_info_get_name (cinfo));
}

string AlsaPatchBay::getAddrName (const Addr& addr) const {
  return getClientName (&addr) + "::" + getPortName (&addr);
}

void AlsaPatchBay::report (DebugLevel level, string message) {
  if (level <= debugLevel) {
    cout << message << endl;
  }
}

void AlsaPatchBay::reportLog (DebugLevel level, string message) {
  report (level, message);
  if (level <= debugLevel) {
    if (_log) {
      _log->value ((message + "\n" + _log->value()).c_str());
    }
  }
}

void AlsaPatchBay::alsaInit () {
  int err;

  err = snd_seq_open(&seq, "default",
                     SND_SEQ_OPEN_DUPLEX,
                     SND_SEQ_NONBLOCK);
  if (err) {
    string s ("couldn't open alsa sequencer");
    AlsaPatchBay::report (ERROR, s);
    throw Exception (s);
  }
}

void AlsaPatchBay::getSubscriptions () {
//  APB_REPORT (DEBUG4, string("looking for subscriptions"));
  _subscriptions.clear ();

  snd_seq_query_subscribe_t * subsinfo;
  snd_seq_query_subscribe_alloca (&subsinfo);

  for (list<Addr>::iterator readi = _readPorts.begin ();
       readi != _readPorts.end ();
       ++readi) {
    snd_seq_query_subscribe_set_root(subsinfo, &(*readi));
//    snd_seq_query_subscribe_set_type(subsinfo, SND_SEQ_QUERY_SUBS_READ);
    snd_seq_query_subscribe_set_index (subsinfo, 0);
    while (!snd_seq_query_port_subscribers (seq, subsinfo)) {
      snd_seq_addr_t addr = *snd_seq_query_subscribe_get_addr (subsinfo);
      _subscriptions.push_back (Subscription(*readi, addr));
      reportLog (DEBUG4, string ("Found subscription ") + getAddrName(*readi)
                         + " -> " + getAddrName (addr));
      
      snd_seq_query_subscribe_set_index (subsinfo,
        snd_seq_query_subscribe_get_index (subsinfo) + 1);
    }
  }

  // FIXME: RAH: Register with the system announcer
}

void AlsaPatchBay::getPorts () {
  _readPorts.clear();
  _writePorts.clear();

  snd_seq_client_info_t * cinfo;
  snd_seq_client_info_alloca (&cinfo);
  snd_seq_client_info_set_client (cinfo, -1);

  snd_seq_port_info_t * pinfo;
  snd_seq_port_info_alloca (&pinfo);
  
  
  while (snd_seq_query_next_client (seq, cinfo) >= 0) {

    snd_seq_port_info_set_client(pinfo, snd_seq_client_info_get_client(cinfo));
    snd_seq_port_info_set_port (pinfo, -1);

    while (snd_seq_query_next_port (seq, pinfo) >= 0) {

      int caps = snd_seq_port_info_get_capability (pinfo);

      if (caps & SND_SEQ_PORT_CAP_READ ||
          caps & SND_SEQ_PORT_CAP_DUPLEX) {
        reportLog (DEBUG4, string("Found read port ") +
                   getAddrName (*snd_seq_port_info_get_addr (pinfo)));

        _readPorts.push_back (*snd_seq_port_info_get_addr (pinfo));
      }

      if (caps & SND_SEQ_PORT_CAP_WRITE ||
          caps & SND_SEQ_PORT_CAP_DUPLEX) {

        reportLog (DEBUG4, string("Found write port ") +
                   getAddrName (*snd_seq_port_info_get_addr (pinfo)));

        _writePorts.push_back (*snd_seq_port_info_get_addr (pinfo));
      }

    }
  }
  
}

AlsaPatchBay * AlsaPatchBay::_apb (0);

void AlsaPatchBay::init (int argc, char ** argv) {
  _apb = new AlsaPatchBay (argc, argv);
}

void AlsaPatchBay::run () {
  _apb->runI ();
}

void AlsaPatchBay::runI () {
  refreshPortButtons ();
  getSubscriptions ();
  Fl::run();
}

} /* namespace APB */

int main(int argc, char ** argv) {
  using namespace APB;
  APB::AlsaPatchBay::init (argc, argv);
  APB::AlsaPatchBay::run ();
  return 0;
}