/*
 * ALSA Patch Bay
 *
 * Copyright (C) 2002 Robert Ham (node@users.sourceforge.net)
 *
 * You have permission to use this file under the GNU General
 * Public License, version 2 or later.  See the file COPYING
 * for the full text.
 *
 */

#include <iostream>
#include <list>
#include <string>
#include <wchar.h>
#include <alsa/asoundlib.h>
#include <FL/Fl.H>
#include <FL/Fl_Window.H>
#include <FL/Fl_Multiline_Output.H>
#include <FL/Fl_Button.H>
#include <FL/Fl_Scroll.H>
#include <FL/Fl_Pack.H>
#include <FL/Enumerations.H>
#include <FL/fl_draw.H>

#define APB_REPORT(level, message) \
  AlsaPatchBay::report (\
    level, \
    string(__FILE__) + ":" + Misc::int2string(__LINE__) + ":" + \
      __PRETTY_FUNCTION__ + ": " + message)

#define DEBUG_STRING string(__FILE__) + ":" + Misc::int2string(__LINE__) + ":" + \
      __PRETTY_FUNCTION__ + ": "

namespace APB {

typedef enum { ERROR = 0,
               WARNING,
               NOTICE,
               DEBUG,
               DEBUG1,
               DEBUG2,
               DEBUG3,
               DEBUG4,
               DEBUG5 } DebugLevel;

typedef snd_seq_addr_t Addr;
typedef pair<Addr, Addr> Subscription; /* (source, dest) */


struct Exception {
  string _desc;
  Exception (string desc) : _desc (desc) {}
};

class Misc {
  public:
    static string int2string (int);
    static bool addrCmp (const Addr&, const Addr&);
};

Fl_Color getColor () {
  static int c = -1;
  ++c;
  if (c > (int) FL_WHITE) c = (int) FL_BLACK;
  return (Fl_Color) c;
}

class PortButton : public Fl_Button {
  private:
    Addr _addr;
    string _label;
    Fl_Color _color;
  public:
    PortButton (int x, int y, int w, int h, const Addr& addr, const string& label)
      : _label (label), Fl_Button (x, y, w, h, label.c_str()), _addr (addr), _color (getColor()) {}
    void addr (const Addr& addr) { _addr = addr; }
    Addr addr () { return _addr; }
    Fl_Color color () { return _color; };
};

class AlsaPatchBay : public Fl_Window {
  private:

    static AlsaPatchBay * _apb;

    Fl_Button * _selectedButton;
    Addr _selectedPort;
    
    // fltk variables
    Fl_Multiline_Output * _log;
    Fl_Button * _quitButton;
    Fl_Button * _refreshButton;
    list<PortButton *> _readButtons;
    list<PortButton *> _writeButtons;
    Fl_Scroll * _scrollPane;
    Fl_Pack * _readPack;
    Fl_Pack * _writePack;
    
    // fltk functions
    void fltkInit (int argc, char ** argv);
    static void quitButtonPushed (Fl_Widget * widget, void * data);
    static void refreshButtonPushed (Fl_Widget * widget, void * data);
    void refreshButtonPushed ();
    void refreshPortButtons ();
    void refreshSubscriptions ();
    static void readButtonPushed (Fl_Widget * widget, void * data);
    void readButtonPushed (Fl_Button * button, const Addr * const addr);
    static void writeButtonPushed (Fl_Widget * widget, void * data);
    void writeButtonPushed (PortButton * button, const Addr& addr);
    
    PortButton * getReadButton (const Addr& addr);
    PortButton * getWriteButton (const Addr& addr);
    PortButton * getButton (const Addr& addr,
                            const list<PortButton *>& portButtons);
    // alsa variables
    snd_seq_t * seq;
    list<Addr> _readPorts;
    list<Addr> _writePorts;
    list<Subscription> _subscriptions;
    
    
    // alsa functions
    void alsaInit ();
    void getPorts ();
    void getSubscriptions();
    static void alsaErrorHandler(const char *file,
                                 int line,
                                 const char *function,
                                 int err,
                                 const char *fmt, ...);
    string getPortName (const Addr * const addr) const;
    string getClientName (const Addr * const addr) const;
    string getAddrName (const Addr& addr) const;
    void unsubscribe (const Addr& from, const Addr& to);
  protected:
    virtual void draw ();
  public:
    static const unsigned short buttonWidth = 70;
    static const unsigned short buttonHeight = 24;
    static void report (DebugLevel, string);
    void reportLog (DebugLevel, string);
    static void init (int argc, char ** argv);
    static void run ();
    AlsaPatchBay (int argc, char ** argv) :
      Fl_Window (700, 400, "ALSA Patch Bay"), _selectedButton (0) {
      if (_apb) {
        throw Exception (string ("an AlsaPatchBay already exists"));
      }
      alsaInit ();
      fltkInit (argc, argv);
      
    }
    void runI ();
    void stopI ();
};

} /* namespace APB */
